# Copyright (C) 2006 Kumar McMillan
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

import os
import sys

from testtools.config import debug
from testtools.storage import *
from testtools.exceptions import *
from testtools.fixtures.registry import LoadMap
from testtools.fixtures import registry
from warnings import warn
from copy import copy

class Fixture(object):
    """Encapsulates one or more datasets to be loaded into a storage medium.
    
    Most likely you will only use one of the wrapper classes 
    designed for dealing with a specific storage medium
    (i.e. `SOFixture` for SQLObjects or `CsvFixture` for csv files).
    
    before looking at how you load data and work with it, here 
    is how fixtures are designed to work inside a test case::
        
        >>> from testtools.testself.myfixtures import EmployeeData
        >>> class TestTPSCoverSheet:
        ...     def setUp(self):
        ...         self.fxt = EmployeeData()
        ...     def tearDown(self):
        ...         self.fxt.clean()
        >>> 
    
    Keyword Arguments
    -----------------
    
    - data -- if this is not None then your data will be loaded from here. \ 
                see Defining Data. Setting this keyword will also make your \ 
                object unregistrable.
    - stor_obj -- object to store data.  if you don't provide this here or \
                in  meta, you will get a ValueError. 
    - clean -- if this is True, your object is cleanable, meaning that its \ 
                data can be deleted from the storage object (i.e. when db \
                load fails, all will be deleted for you in try/except). \ 
                Defaults to False.
    - build_meta -- if this is True then the internal meta object will be \ 
                built from the outermost parent class (defaults to True). \ 
                This is only provided so that wrapper classes can \
                do their own building, which was necessary for some reason.
    - is_registrable -- if True (the default) the fixture object will have \
                its data loaded only once per class instance unless data is \ 
                unloaded via `Fixture.clean`().  This is set internally \
                for file-based fixtures (i.e. `CsvFixture`) and the like.
    
    Now for the more low level examples ...
    
    Example with `CsvFixture`
    -------------------------

        >>> from testtools.fixtures import CsvFixture
        >>> class Products(CsvFixture):
        ...     class meta:
        ...         data = ( ('product_on_sale', {'name':'toy truck', 'on_sale':True}), )
        >>> fxt = Products(filename='/tmp/testtools_example.csv' )
        >>> fxt.product_on_sale
        <class 'Fixture.Values' dict(on_sale='True', name='toy truck') from module testtools.fixtures.fixtures>
    
    
    Defining data
    -------------
    Data is an iterable of tuples where each row unpacks to a context 
    key and dictionary of values.  The context key becomes an attribute 
    and/or dictionary key in the fixture object.
    
    For example:
    
        >>> # NOTE: the trailing comma here forces the type into a tuple :
        >>> fxt.load( (('reversible_slacks_on_sale', 
        ...             {'name': 'slacks, reversible', 'on_sale': True}),) )
        >>> fxt.reversible_slacks_on_sale
        <class 'Fixture.Values' dict(on_sale='True', name='slacks, re...') from module testtools.fixtures.fixtures>
    
    A fixture's data can also be set in the following ways, respectively :
    
    From the data keyword :
    
        >>> from testtools.testself.model import Employee
        >>> from testtools.fixtures import SOFixture
        >>> fxt = SOFixture(data = (
        ...     ('bob', {'name':'bob','hair_color':'black'}),), 
        ...                                     so_class=Employee)
        >>> fxt.bob
        <class 'Fixture.Values' dict(name='bob', hair_color='black') from module testtools.fixtures.fixtures>
    
    from meta.data :
    
        >>> # here the SQLObject "Employee" is pulled out from meta.env 
        >>> # because it matches the class name employee
        >>> from testtools.testself import model
        >>> class Employee(SOFixture):
        ...     class meta:
        ...         env = model
        ...         data = (('jane', {'name':'jane','hair_color':'brown'}),)
        >>> fxt = Employee()
        >>> fxt.jane
        <class 'Fixture.Values' dict(name='jane', hair_color='brown') from module testtools.fixtures.fixtures>
    
    ... or by defining a data() method :
    
        >>> # or you can set meta.so_class to the specific SQLObject
        >>> class JaneEmployee(SOFixture):
        ...     class meta:
        ...         so_class = model.Employee
        ...     def data(self):
        ...         return (('jane', {'name':'jane','hair_color':'brown'}),)
        >>> fxt = JaneEmployee()
        >>> fxt.jane
        <class 'Fixture.Values' dict(name='jane', hair_color='brown') from module testtools.fixtures.fixtures>
    
    A Word About Meta
    -----------------
    
    `Fixture.meta` is a namespace for common properties and is populated
    with any keyword arguments (above) that might be passed in to an instance.
    Most `Fixture` wrappers, like `SOFixture`, use meta to 
    set up some common attributes for you.  With the exception of 
    "build_meta" all default meta attributes are defined like the Keyword 
    Arguments, documented above.
    
    I.E. all instances of EmployeeData will share the same data and 
    will only be loaded once ... but DepartmentData instances have their 
    own data.  
    
    """
    class meta:
        """namespace for `Fixture` properties.
        
        Note that this class is automatically de-referenced within instances
        so you can safely create a class that has its own meta properties 
        like so:
        
            >>> class MyFixture(Fixture):
            ...     class meta:
            ...         clean = True
            >>> 
        
        """
        pass
        
    _reserved_meta_names = ('req')
        
    class RequiredPool(object):
        """encapsulates other fixtures required by a Fixture."""
        def __init__(self):
            object.__setattr__(self, '_pool', [])
            object.__setattr__(self, '_pool_lookup', {})
            
        def __getattr__(self,k):
            if not self._pool_lookup.has_key(k):
                raise AttributeError, (
                    "%s object has no attribute %s" % 
                                (self.__class__.__name__, k))
            return self._pool[ self._pool_lookup[k] ]
            
        def __setattr__(self,k,v):
            if not issubclass(v.__class__, Fixture):
                raise TypeError, (
                    "cannot set attribute '%s' in required pool "
                    "because it is not a Fixture" % k)
            self._pool.append(v)
            self._pool_lookup[k] = len(self._pool)-1
            
    class Values:
        """encapsulates a Fixture's values (dataset)."""
        def __init__(self, **kwds):
            self.__dict__ = kwds
        def __getitem__(self, k):
            return self.__dict__[k]
        def __repr__(self):
            try:
                mod = self.__module__
            except AttributeError:
                mod = '[unknown]'
                
            def shorten(s):
                s = str(s)
                if len(s) > 10:
                    return s[0:10] + '...'
                else:
                    return s
                
            pdict = ("dict(" + ", ".join( [
                "%s='%s'" % (k,shorten(v)) for k,v in self.__dict__.items()] ) 
                                + ")")
            return ("<class 'Fixture.%s' %s from module %s>" % 
                            (self.__class__.__name__, pdict, mod))
        __str__ = __repr__
        
        def items(self):
            return self.__dict__.items()
    
    def __init__(self,  data=None, build_meta=True, 
                        clean=False, is_registrable=True):
        """see `Fixture` for signature."""
        self._use_registry = (data is None and is_registrable)
        if debug.stream:
            id = LoadMap.ID(self)
            id_name = self.__class__.__name__
        
        preloaded = False
        if self._use_registry:
            if LoadMap.is_registered(self):
                if debug.stream:
                    print >>debug.stream, (
                        "%s(): POPULATE self %s from registry; RETURN" % 
                                                            (id_name, id))
                LoadMap.populate(self)
                preloaded = True
            else:
                if debug.stream:
                    print >>debug.stream, (
                                "%s(): UNREGISTERED : %s" % (id_name, id))
        else:
            if debug.stream:
                print >>debug.stream, "%s(): NOT registrable" % id_name
        
        if build_meta:
            self._build_meta(dict(data = None, stor_obj = None, clean=clean))
        
        if not preloaded:
            self.meta.req = Fixture.RequiredPool()
            self.meta._loaded = [] # ref to all StorObj
            self.meta._is_loaded = False
            self.meta._dataset = None # ref to last dataset loaded
            self.meta._length = 0 # length of datasets
        
        def load_data(data, clean=clean):
            if clean:
                self.clean()
            self.load(data=data)
        
        if data is not None:
            if debug.stream:
                print >>debug.stream, (
                            "%s(): LOAD data from data keyword" % id_name)
            load_data(data)
        else:
            if not preloaded:
                if not self.meta.data:
                    data_by_method = self.data()
                    if data_by_method is not None:
                        if debug.stream:
                            print >>debug.stream, (
                                "%s(): LOAD data from self.data()" % id_name)
                        load_data(self.data())
                else:
                    if debug.stream:
                        print >>debug.stream, (
                                "%s(): LOAD data from self.meta.data" % 
                                                                id_name)
                    load_data(self.meta.data)
    
    def __len__(self):
        return self.meta._length
    
    def __repr__(self):
        try:
            mod = self.__module__
        except AttributeError:
            mod = '[unknown]'
            
        pkeys = str(tuple( [k for k,v in self.items()] ))
        max = 65
        if len(pkeys) > max:
            pkeys = pkeys[0:max] + '...)'
        
        return "<class '%s' keys=%s from module %s>" % \
                        (self.__class__.__name__, pkeys, mod)
    __str__ = __repr__
    
    def _build_meta(self, defaults):
        """set default meta vars or merge in the defaults.
        
        """
        if self.meta is self.__class__.meta:
            # de-ref it :
            class custom_meta(self.__class__.meta):
                pass
            self.meta = custom_meta
            
        for k,v in defaults.items():                
            if not hasattr(self.meta, k):
                setattr(self.meta, k, v)
                
        # don't accept unknown meta vars:
        for k in dir(self.meta):
            if (not k.startswith('_') and 
                    k not in Fixture._reserved_meta_names):
                if k not in defaults:
                    raise TypeError, (
                        "%s() got unknown meta attribute or keyword '%s'" %
                            (self.__class__.__name__, k))
    
    def _ismeta(self, name):
        """true if this value is meta and should be ignored."""
        return name == 'meta'
    
    def _new_stor_obj(self):
        return self.meta.stor_obj()
    
    def add_fixture(self, fixture):
        """merges attributes from another fixture into self."""
        for k,v in fixture.items():
            self.__dict__[k] = v
    
    def assert_has(self, dataset, transkey=None, transval=None, quiet=False):
        """assert this fixture has a dataset somewhere in it.
        
        fetching dataset items will be tried like:
        dataset.key and dataset[key]
        
        Keyword arguments
        ------------------
        - transkey -- dict called like: \
            ''dataset_key = transkey[fixture_key]'' \
            to match dataset items with fixture items
        - quiet -- True if you don't want a big explanation of mismatches
        
        """
        any_matched = False
        mismatched = {}
        
        for context, fxt in self.items():
            looked = False
            found_one = True
            
            for fkey, fxt_val in fxt.items():
                dkey = fkey
                if transkey is not None:
                    try:
                        dkey = transkey[dkey]
                    except (KeyError, TypeError):
                        pass
                try:
                    dval = getattr(dataset, dkey)
                except AttributeError:
                    try:
                        dval = dataset[dkey]
                    except (KeyError, TypeError), e:
                        if transkey is None:
                            raise
                        else:
                            raise KeyError(
                            '%s not a key in dataset (transkey=%s)' % 
                                                    (str(e), transkey))
                
                if callable(transval):
                    dval = transval(dval)
                    fxt_val = transval(fxt_val)
                if fxt_val != dval:
                    if not quiet:
                        mismatched.setdefault(context,[])
                    
                        def explain(val):
                            return (callable(transval) and '%s(%s)' % 
                                            (transval.__name__, val) or val)
                        mismatched[context].append('%s != %s' % 
                                            (explain(fxt_val), explain(dval)))
                    found_one = False
                looked = True
                
            if looked and found_one:
                any_matched = True
                
        if not any_matched:
            raise AssertionError(   'dataset (%s) did not match any '
                                    'fixture datasets%s' % 
                                        (dataset, not quiet and (' (%s)' % 
                                                        mismatched) or ''))
        
    def clean(self, cascade = True):
        """perform clean() on all storage objects loaded for 
        this instance only.
        
        goes in reverse order so foreign key constraints aren't violated.
        *does not* clean self's fixture attributes.
        
        NOTE: clean() raises an error if self.meta.clean is False.  
        See `Fixture` doc for more info on meta.clean
        
        """
        if debug.stream:
            print >>debug.stream, ('%s.clean(): START %s' % 
                                        (self.__class__.__name__, 
                                            LoadMap.ID(self)))
            
        if (not hasattr(self.meta,'clean') or 
                (hasattr(self.meta,'clean') and not self.meta.clean)):
            # I should probably think about this more ...
            raise TypeError, (
                "'%s' is uncleanable (self.meta.clean was False)" % 
                                            self.__class__.__name__)
                                
        if not self.meta._is_loaded:
            if debug.stream:
                print >>debug.stream, (
                            '%s.clean(): NOT LOADED; exiting clean()' % 
                                                (self.__class__.__name__))
            return
            
        def clean_obj(obj, prompt=''):
            if debug.stream:
                print >>debug.stream, (
                        '%s.clean(): %s %s' % 
                            (self.__class__.__name__, prompt, obj))
            obj.clean()
            
        # need to clean in reverse order for the
        # foreign keys ... 
            
        ## any *storage* objects that are loaded ...
        self.meta._loaded.reverse()
        try:
            if hasattr(self.meta, 'clean') and self.meta.clean:
                for o in self.meta._loaded:
                    clean_obj(o, prompt='CLEANING StorObj')
        finally:
            self.meta._loaded.reverse() # put it back.  hmmm.
            
        self.meta._is_loaded = False
        
        ## any *fixtures* that were required ...
        if debug.stream:
            print >>debug.stream, ('%s.clean(): REQUIRES %s' % 
                    (   self.__class__.__name__, 
                        [o.__class__.__name__ for o in self.meta.req._pool]))
        if cascade:
            try:
                for o in self.meta.req._pool:
                    if hasattr(o.meta, 'clean') and o.meta.clean:
                        clean_obj(o, prompt='CLEANING REQUIRED')
            finally:
                pass
        else:
            if debug.stream:
                print >>debug.stream, (
                "%s.clean(): NO CASCADE: required fixtures not cleaned" %                     
                                                (self.__class__.__name__))
            
        LoadMap.unregister(self)
    
    def close(self):
        [o.close() for o in self.meta._loaded]
    
    def data(self):
        """hook for subclasses to return a data definition.
        
        returns a tuple of tuples.
        see class signature for example of the format
        
        """
        return None
    
    def items(self):
        """yields (fixture_key, fixture) in self."""
        for (k,v) in self.__dict__.items():
            if k.startswith('_') or self._ismeta(k) or callable(v):
                continue
            yield (k,v)
        
    def load(self, data=None, stor_obj=None, clean=False):
        """load all data using stor_obj.save().
        
        data is an iterable of context and dataset that looks like this::
            (('context1', {} ), ('context2', {} ), ... )
        
        Each dataset is stored in self as an attribute (or key) named context.
        For example, self.context1 == {}.
        if clean is True, calls stor_obj.clean() *before* loading.
        
        """
        if data is None:
            data = self.meta.data
        if data is None:
            raise ValueError("data and self.meta.data cannot be None")
        
        self.meta._dataset = data
        
        if stor_obj is None:
            stor_obj = self._new_stor_obj()
            
        if clean:
            stor_obj.clean()
            
        for context, dataset in data:
            if type(dataset) != type({}):
                raise ValueError(
                "fixture dataset at context '%s' must be a dictionary" % 
                                                                    context)
            if self.__dict__.has_key(context):
                if self._ismeta(context):
                    raise AttributeError, ( 
                        "'meta' is a reserved word and cannot "
                        "be used as a fixture context")
                else:
                    raise AttributeError, (
                        "a fixture with named context '%s' has "
                        "already been loaded" % context)
                        
            v = Fixture.Values(**dataset)
            self.__dict__[context] = v
        
            stor_obj.save(**dataset)
            if self._use_registry:
                LoadMap.update(self)
                
            self.meta._length += 1
            
        self.meta._loaded.append(stor_obj)
        self.meta._is_loaded = True

class FixtureCollection(Fixture):
    """returns a collection of Fixture instances, all attributes shared.
    
    accepts fixture instances as arguments, returns an instance with all 
    attributes combined.
    
    """
    def __init__(self, *fixtures):
        
        self._fixture_coll = list(fixtures)
        try:
            for f in self._fixture_coll: self.add_fixture( f )
        except Exception, err:
            tb = sys.exc_info()[2]
            try:
                self.clean()
            except Exception, e:
                try:
                    if debug.stream:
                        print >>debug.stream, (
                            "%s() INIT aborted by %s" % 
                            (self.__class__.__name__, e.__class__.__name__))
                except:
                    pass
            raise err, None, tb
        
        self.close()
    
    def __len__(self):
        # sum of all lengths ...
        return reduce(lambda x,y: x+y, [len(f) for f in self._fixture_coll] )
        
    def clean(self, cascade=True):
        self._fixture_coll.reverse()
        try:
            for f in self._fixture_coll: 
                f.clean(cascade = cascade)
        finally:
            self._fixture_coll.reverse()
        
    def close(self):
        for f in self._fixture_coll: f.close()
        
    def load(self, *arg,**kw):
        raise NotImplementedError
        
affix = FixtureCollection

def with_fixtures(*fixtures, **kw):
    """decorator that attaches fixtures to a test def.
    
    Positional args are `Fixture` classes to be instantiated 
    and passed into `affix`
    
    The resulting `FixtureCollection` object will be the first 
    argument to your decorated def
    
    Keyword Args:
    -------------
    - clean -- when True (default), all fixtures will be cleaned in a \
            finally block
    
    For example :
    
    >>> from testtools.fixtures import CsvFixture, with_fixtures
    >>> class MyBooks(CsvFixture):
    ...     class meta:
    ...         filename = open('/tmp/testtools_example.csv', 'w')
    ...         data = (('lolita', {'title':'lolita'}),)
    ...         clean = True
    ... 
    >>> @with_fixtures(MyBooks)
    ... def test_with_books(fxt):
    ...     print fxt.lolita
    ...     # test something ...
    ... 
    >>> test_with_books()
    <class 'Fixture.Values' dict(title='lolita') from module testtools.fixtures.fixtures>
    >>> 
    
    """
    if not kw.has_key('clean'):
        kw['clean'] = True
    clean = kw['clean']
    
    def with_fixtures_deco(_def):
        def _call_def(*a,**kw):
            fxt = affix(*[f() for f in fixtures])
            fxt.close()
            try:
                _def(fxt, *a, **kw)
            finally:
                if clean:
                    fxt.clean()
        _call_def.__name__ = _def.__name__
        return _call_def
    return with_fixtures_deco
    