# Copyright (C) 2006 Kumar McMillan
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

"""Private registry for the Fixture class.

"""
from copy import copy
from testtools.config import debug

def unregister_fixtures():
    """remove any and all registered fixtures from the registry.
    
    NOTE that fixtures are not actually unloaded (see `Fixture.clean` )
    """
    LoadMap.clear()

class LoadMap(object):
    """registry of loaded fixtures.
    
    loaded fixtures are registered by the *class* of their instance.
    If you subclass a fixture it would be considered a completely new fixture.
    
    THOUGHTS: Do we need a way to explicitly say which class should be 
    considered our key?  I don't think so but if that's the case we could 
    look for a custom __fixture_id__ attribute or something
    
    """
    class ID(object):
        def __init__(self, obj):
            self.instance_id = id(obj)
            klass = obj.__class__
            self.id = id(klass)
            self.module = klass.__module__
            self.name = klass.__name__
        
        def __repr__(self):
            return ("<class '%s' ID %s (instance ID %s) from module %s>" % 
                            (self.name, hex(self.id), 
                             hex(self.instance_id), self.module))
    
    registry = {}
    
    def __init__(self):
        raise NotImplementedError, (
                "you cannot create an instance of the registry")
    
    def clear(self):
        self.registry = {}
    clear = classmethod(clear)
    
    def populate(self, fixture):
        """populates fixture instance with loaded fixture. 
        
        returns fixture.
        """
        id = self.ID(fixture).id
        if self.is_registered(id = id):
            for k,v in self.registry[id].data.items():
                fixture.__dict__[k] = copy(v)
            
            fixture.meta = self.registry[id].meta
            
        return fixture
    populate = classmethod(populate)
    
    def is_registered(self, fixture=None, id=None):
        """return True if this fixture is registered as loaded"""
        if id is None:
            if fixture is None:
                raise TypeError, (
                        "fixture cannot be None (incorrect arg count)")
            id = self.ID(fixture).id
        return self.registry.has_key(id)
    is_registered = classmethod(is_registered)
    
    def register(self, fixture, id=None):
        """registers a fixture that has been loaded."""
        if id is None:
            id = self.ID(fixture).id
            
        class storage: pass
        self.registry[id] = storage()
        self.registry[id].data = dict([(k,v) for k,v in fixture.items()])
        self.registry[id].meta = fixture.meta
    register = classmethod(register)
    
    def update(self, fixture):
        """updates registered fixture's loaded data."""
        id = self.ID(fixture).id
        if not self.is_registered(id = id):
            self.register(fixture, id=id)
        else:
            self.registry[ id ].data.update(
                    dict([(k,v) for k,v in fixture.items()]))
            self.registry[ id ].meta = fixture.meta
    update = classmethod(update)
    
    def unregister(self, fixture):
        """tell registry that this fixture has been 
        unloaded (Fixture.clean()).
        """
        id = self.ID(fixture).id
        if self.is_registered(id = id):
            del self.registry[ id ]
    unregister = classmethod(unregister)
    