# Copyright (C) 2006 Kumar McMillan
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

"""Storable `Fixture` objects.

In other words : `Fixture` descendants to work with specific `Stor` objects.

"""

from testtools.fixtures import Fixture
from testtools.exceptions import *
from testtools.storage import *
from testtools.config import debug

class CsvFixture(Fixture):
    """Fixture storable as a CSV file.
    
    The following are attributes you can set as keywords or meta attributes 
    (see `Fixture` for examples on configuring a class using meta)
    
    - addfields -- if True first row in the CSV file will contain a field list
    - data -- dataset(s) to load (see `Fixture` for format)
    - encoding -- string encoding to use when writing to file
    - fields -- field list (if None, fields will be keys of the first \
        data item)
    - filename -- name of file or filelike object
    - dialect -- csv dialect to use (default: excel)
    
    """
    
    def __init__(self, **kw):
        self._build_meta(dict(  addfields = False,
                                data = None,
                                encoding = None,
                                fields = None,
                                filename = None,
                                stor_obj = CsvStor,
                                clean = False,
                                dialect = 'excel'))
        local_kw = {'data':None}
        
        for k,v in kw.items():
            # anything inputted (except locals), override the 
            # meta defaults ...
            if not local_kw.has_key(k):
                setattr(self.meta, k, v)
        
        for k,v in local_kw.items():
            if not kw.has_key(k):
                kw[k] = v
        
        super(CsvFixture, self).__init__(   
                data=kw['data'], clean=self.meta.clean, 
                build_meta=False, is_registrable = False)
    
    def _new_stor_obj(self):
        if self.meta.fields is None:
            self.meta.fields = self.meta._dataset[0][1].keys()
            
        return self.meta.stor_obj(  
                self.meta.filename, self.meta.fields, 
                clean = self.meta.clean, addfields = self.meta.addfields, 
                encoding = self.meta.encoding, dialect = self.meta.dialect)

def find_so_from_data(meta, fixture_class):
    """finder method for SOFixture that transforms 
    SomeFixtureData into the name SomeFixture.
    
    """
    return fixture_class[0:-4]

def find_so_from_fixture(meta, fixture_class):
    """finder method for SOFixture that passes through 
    your fixture name (no transformation).
    
    This is the default.
    
    """
    return fixture_class

def find_so_from_table(meta, fixture_class):
    """finder method for SOFixture that transforms your 
    fixture class as if it were a table name. 
    
    It assumes your fixture class is named exactly like your database 
    table name and its SQLObject should be translated like it would in 
    SQLObject land.
    
    i.e. fixture class products_to_offers becomes ProductsToOffers
    
    """
    # use sqlobject.main.findClass ?? had trouble with that
    def part_to_attr(part):
        if part == 'id':
            return part.upper()
        else:
            return part.capitalize()
    return ''.join([part_to_attr(part) for part in fixture_class.split('_')])
    
class SOFixture(Fixture):
    """Fixture storable by an SQLObject.
    
    The following are attributes you can set as keywords or meta attributes 
    (see `Fixture` for examples on configuring a class using meta)
    
    - data -- datasets to load (see `Fixture` for format)
    - so_class -- default SQLObject class to use \
                (if missing, `SOFixture.find_so_class` will be consulted)
    - env -- dict or something with attributes 
    - finder -- callable to get so_class from self.__class__ \
                (default is `find_so_from_fixture` )
    - clean -- if True, object is cleanable (defaults to False)
    - autocommit -- if False (default) storage is performed in a \
            transaction.  see `SOStor`.
    
    """
    
    def __init__(self, data=None, so_class=None, clean=False, 
                        autocommit=False ):        
        self._build_meta(dict(  
                data=None, env={}, stor_obj=SOStor, so_class=so_class, 
                finder=find_so_from_fixture, 
                clean=clean, autocommit=autocommit))
        
        if not self.meta.so_class:
            self.meta.so_class = self.find_so_class()
        
        super(SOFixture, self).__init__(data=data, build_meta=False )
    
    def find_so_class(self, fixture_class=None):
        """searches meta.env for an SQLObject named by meta.finder.
        
        meta.env is a dict or something with attributes.
        meta.finder is a callback that accepts the fixture class name of 
        this instance.
        returns SQLObject class.
        
        """
        
        if fixture_class is None:
            fixture_class = self.__class__.__name__
            
        # hmmm callback wants an instance of meta ??
        expected_so_class = self.meta.finder(self.meta(), fixture_class)
        
        so_class = None
        if hasattr(self.meta.env, expected_so_class):
            so_class = getattr(self.meta.env, expected_so_class )
        else:
            try:
                so_class = self.meta.env[expected_so_class]
            except (KeyError,TypeError), exc:
                pass
        if so_class is None:
            raise LoadError(
                "expected to get attribute or key '%s' as SQLObject "
                "from self.meta.env, finder=%s" % 
                        (expected_so_class, self.meta.finder))
                        
        return so_class
    
    def _new_stor_obj(self):
        return self.meta.stor_obj(
                self.meta.so_class, autocommit = self.meta.autocommit)
        