# Copyright (C) 2006 Kumar McMillan
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

"""IO-related testing tools.

An example ...
    
    >>> from testtools.io import TempIO, readfile
    >>> from os.path import exists, join
    >>> tmp = TempIO()
    >>> tmp.putfile('data.txt', 'lots of nonsense')
    >>> print readfile(join(tmp.root, 'data.txt'))
    lots of nonsense
    >>> assert exists(tmp.newdir('incoming'))
    >>> assert exists(tmp.incoming)
    >>> del tmp ## or let it destruct naturally

"""

import os
from os.path import join, exists, split, basename
from tempfile import mkdtemp
import atexit

_tmpdirs = {}

def _expunge(tmpdir):
    """called internally to remove a tmp dir."""
    global _tmpdirs
    
    if exists(tmpdir):
        import shutil
        shutil.rmtree(tmpdir)
    
        if _tmpdirs.has_key(tmpdir):
            del _tmpdirs[tmpdir]
        
def _expunge_all():
    """exit function to remove all registered tmp dirs."""
    global _tmpdirs
    [ _expunge(d) for d,id in _tmpdirs.items() ]
    
# this seems to be a safer way to clean up since __del__ can
# be called in a volatile environment :
atexit.register(_expunge_all)

def mkdirall(path, mkdir=os.mkdir):
    """walks the path and makes any non-existant dirs.
    
    optional keyword `mkdir` is the callback for making a single dir
    
    """
    if path[-1] == os.path.sep: 
        path = path[0:-len(os.path.sep)] # trailing slash confused exists()
        
    root = path[0] == os.path.sep and os.path.sep or ''
    paths = split(path)[0].split(os.path.sep)
    if len(paths):
        accum = ''
        for p in paths:
            if p is '':
                continue # slash prefix will cause this
            accum = join(accum, p)
            abs = join(root, accum)
            if not exists(abs): mkdir(abs)
    
    mkdir(path)

def readfile(filename, mode='rU', pad=False):
    """debug util that returns file contents.
    
    NOTE: does a *full* read on filename, you might not want to do that.
    
    if pad == True then contents will be padded with 
    --begin-- and --end-- blocks.
    
    returns contents as a string.
    """
    f = open(filename, mode)
    contents = f.read()
    f.close()
    
    if pad:
        return "\n".join(('--begin--',contents,'--end--'))
    else:
        return contents

def putfile(filename, contents, filelike=None, mode=None):
    """opens filename in writing mode, writes contents and closes.
    
    if filelike is None then it will be created with open() and the
    prefixed path will be walked to create non-existant dirs.
    
    """
    if mode is None:
        mode = 'w'
    if filelike is None:
        parent = split(filename)[0]
        if parent and not exists(parent):
            mkdirall(parent)
        filelike = open(filename, mode)
        
    filelike.write(contents)
    filelike.close()
    
class TempIO(object):
    """self-destructable temporary directory root.
    
    Takes the same keyword args as tempfile.mkdtemp.
    
    You will most likely create this in a test module like so 
    (`nosetests`_/ `py.test`_ style) :
    
        >>> tmp = None
        >>> def setup_module(self):
        ...     self.tmp = TempIO()
        >>> def teardown_module(self):
        ...     del self.tmp
        >>> def test_something():
        ...     tmp.root
        ...     # ...
        >>>

    NOTE: due to the unpredictability of when
    destructors get called, you may want to explicitly
    delete your instance in a teardown method.  however, an atexit 
    function will try and clean up too.
    
    .. _py.test: http://codespeak.net/py/current/doc/test.html
    .. _nosetests: http://nose.python-hosting.com/
    
    """
    
    def __init__(self, **kw):
        global _tmpdirs
        if not kw.has_key('prefix'):
            # a breadcrumb ...
            kw['prefix'] = 'tmp_testtools_'
            
        self.root = mkdtemp(**kw)
        _tmpdirs[self.root] = id(self)
        
    def __del__(self):
        """removes the root directory and everything under it.
        
        """
        try:
            _expunge(self.root)
        except:
            # means atexit didn't get it ...
            # this is the last resort.  let this raise an exception?
            # apparently we can even get import errors in __del__,
            # like for shutil.  how very strange.
            pass
    
    def __repr__(self):
        return "<%s '%s' at %s>" % (self.__class__.__name__, 
                                    self.root, hex(id(self)))
    
    def mkdir(self, name, mkall=True):
        """makes a directory in the root and returns its full path.
        
        if mkall is True, will split the path and make each 
        non-existant directory.  returns full path to new directory.
        
        """
        path = os.path.join(self.root, name)
        if mkall:
            mkdirall(path)
        else:
            os.mkdir(path)
        return path
    
    def mkfile(self, fname):
        """makes a filename in root. 
        
        - the path is relative to your `TempIO` root.
        - all subdirectories are created if they don't exist
        - no checking is done as to whether this file exists or not
        
        returns full path to new file.
        
        """
        relpath, fname = split(fname)
        if relpath and not exists(join(self.root, relpath)):
            if relpath.startswith(os.path.sep):
                relpath = relpath[1:]
            self.mkdir(relpath, mkall=True)
            
        return join(self.root, relpath, fname)
    
    def newdir(self, name):
        """makes a directory in the root. 
        
        it also adds name as a property to this instance.  
        returns full path to new directory.
        
        """
        path = self.mkdir(name)
        setattr(self, name, path)
        return path
    
    def putfile(self, fname, contents):
        """puts new file in your `TempIO` root.
        
        see `mkfile` for how fname is handled.
        
        """
        putfile(self.mkfile(fname), contents)
        