
import os
import csv
import sys
import sre
try:
    import cElementTree as ElementTree
except ImportError:
    from elementtree import ElementTree
import unittest

import testtools
from testtools.config import reset_debug, debug, set_debug_stdout
from testtools.exceptions import *
from testtools.fixtures import *
from testtools.fixtures.registry import LoadMap
from testtools.storage import *
from testtools.io import TempIO

from testtools.testself.util import start_csv

# errm, the other fixtures ..
from testtools.testself.myfixtures import   EmployeeData, DepartmentData, ShopData, csvdict
from testtools.testself.model import *

data = (('blue_hair_color_employee', {
                'id':1, 'name':'Fred', 'hair_color':'blue'}),
        ('boring_employee', {
                'id':2, 'name':'Bob', 'hair_color':'gray'}))

TMP = None

def setup_module(self):
    self.TMP = TempIO()
    LoadMap.clear()
    
def teardown_module(self):
    del self.TMP

class TestFixtures(unittest.TestCase):
    fxt = None
                
    def setUp(self):
        LoadMap.clear()
        self.fxt = Fixture()
        Employee.clearTable()
        Department.clearTable()
        Shop.clearTable()
        reset_debug()
    
    def test_fixture_conflict(self):
        try:
            self.fxt.load(( ('same_key',dict(name='foo',hair_color='red')),
                            ('same_key',dict(name='bar',hair_color='blue'))), 
                            SOStor(Employee))
        except AttributeError:
            pass
        else:
            assert False, 'expected exeception'
    
    def test_fixture_conflict_meta(self):
        try:
            self.fxt.load(( ('meta',dict(name='foo',hair_color='red')),), 
                            SOStor(Employee))
        except AttributeError:
            pass
        else:
            assert False, 'expected exeception'

    def test_fixture_xml(self):
        # write some data to make sure it gets cleaned out :
        xmlf = TMP.mkfile('mock_xml_fix.xml')
        f = open(xmlf, 'w')
        f.write('invlaid></XML in your face!')
        f.close()
        
        xml = XmlStor(open(xmlf, 'a'))
        self.fxt.load(data, xml, clean=True)
        xml.close() # mimic affix?
        
        f = open(xmlf,'r')
        genxml = f.read()
        f.close()
        
        try:
            tree = ElementTree.parse(xmlf)
        except SyntaxError:
            print genxml
            raise
            
        def find(attr, tree=tree, xml=xml):
            return [e.text for e in tree.findall('%s/%s' % (xml.item, attr))]
        
        assert (sre.match("<\?xml version=('|\")1\.0('|\")\?>", genxml), 
                    "no encoding (first line) : %s" % genxml.splitlines()[0])
        assert '1' in find('id')
        assert self.fxt.blue_hair_color_employee.name in find('name')
        assert (self.fxt.blue_hair_color_employee.hair_color in 
                find('hair_color'))
        
        assert '2' in find('id')
        assert self.fxt.boring_employee.name in find('name')
        assert self.fxt.boring_employee.hair_color in find('hair_color')
    
    def test_fixtures_items(self):
        fields = ['id', 'name', 'hair_color']
        csvfile = TMP.mkfile('test_fixtures_items.csv')
        
        self.fxt.load(data, CsvStor(open(csvfile, 'a'), fields, 
                        addfields=True))
        
        num = 0
        for k,v in self.fxt.items():
            if k == 'blue_hair_color_employee':
                assert v['name'] == self.fxt.blue_hair_color_employee.name
                assert v.name == self.fxt.blue_hair_color_employee.name
            elif k == 'boring_employee':
                assert v['name'] == self.fxt.boring_employee.name
                assert v.name == self.fxt.boring_employee.name
            else:
                assert False, 'unexpected items: (%s,%s)' % (k,v)
            num += 1
        assert num > 0
    
    def test_len(self):
        self.fxt = affix(EmployeeData())
        count = 0
        for k,v in self.fxt.items():
            count += 1
        self.assertEqual(len(self.fxt), count)
        
    def test_load(self):
        # insert some data to make sure it does not get cleaned:
        Employee(id=5, name='TEST', hairColor='red')
        
        self.fxt = affix(EmployeeData())
        
        not_cleaned = False
        for emp in Employee.select():
            if emp.id == 5:
                not_cleaned = True
                continue
            self.fxt.assert_has(emp, transkey={'hair_color':'hairColor'})
            
        assert not_cleaned
                
    def test_load_clean(self):
        # insert some data to make sure it gets deleted:
        Employee(id=5, name='TEST', hairColor='red')
        
        self.fxt.load(data, SOStor(Employee), clean=True)
        #print self.fxt
        for emp in Employee.select():
            self.fxt.assert_has(emp, transkey={'hair_color':'hairColor'})
    
    def test_load_clean_backwards(self):
        '''Fixtures.clean() needs to clean in reverse, to avoid
        foreign key errors, but sqlite doesn't enforce foreign keys
        so making a fragile mock here .. :( '''
        class mock:
            cleaned = []
        class MockStor(Stor):
            def clean(self):
                mock.cleaned.append(self)
            def close(self): pass
            def save(self): pass
                
        class StorOne(MockStor): pass
        class StorTwo(MockStor): pass
            
        stor_one = StorOne()
        stor_two = StorTwo()
        
        fxt = Fixture(clean = True)
        fxt.load( (('one',{}),), stor_one, clean=False)
        fxt.load( (('two',{}),), stor_two, clean=False)
        fxt.clean()
        
        self.assertEqual( mock.cleaned[0], stor_two)
        self.assertEqual( mock.cleaned[1], stor_one)
        
        # and then for collections ...
        mock.cleaned = []
        stor_one = StorOne()
        stor_two = StorTwo()
        
        fxt1 = Fixture(clean = True)
        fxt1.load( (('one',{}),), stor_one, clean=False)
        fxt2 = Fixture(clean = True)
        fxt2.load( (('two',{}),), stor_two, clean=False)
        
        fxt = affix(fxt1, fxt2)
        fxt.clean()
        self.assertEqual( mock.cleaned[0], stor_two)
        self.assertEqual( mock.cleaned[1], stor_one)
            
    def test_load_clean_uncleanable(self):
        fxt = Fixture(clean = False)
        self.assertRaises(TypeError, fxt.clean)
    
    def test_load_many_remembers_loaded(self):
        #set_debug_stdout()
        
        fxt = EmployeeData()
        self.assertEqual(Employee.select().count(), len(fxt))
        
        fxt.meta.req.dept = DepartmentData()
        self.assertEqual(Department.select().count(), len(fxt.meta.req.dept))
        
        fxt = EmployeeData()
        fxt = EmployeeData()
        fxt.clean()
        
        #reset_debug()
        self.assertEqual(Employee.select().count(), 0)
        self.assertEqual(Department.select().count(), 0)
    
    def test_load_reference(self):
        unregister_fixtures()
        
        # tests for bug where load flag was getting set in the wrong place
        self.fxt = affix(EmployeeData())
        
        for dept in Department.select():
            if dept.id == 1:
                assert (dept.employee == 
                        Employee.get(self.fxt.nerd_employee.id))
                assert dept.name == self.fxt.nerds_department.name
            else:
                raise AssertionError('unexpected row: %s' % dept)
    
    def test_load_then_clean(self):
        
        unregister_fixtures()
    
        self.fxt = affix(EmployeeData(), DepartmentData(), ShopData())
    
        assert Employee.select().count() > 0
        assert Department.select().count() > 0
        assert Shop.select().count() > 0
    
        self.fxt.clean()
    
        assert Employee.select().count() == 0
        assert Department.select().count() == 0
        assert Shop.select().count() == 0
    
    def test_loaded_fixture_multiple(self):
        self.fxt = affix(EmployeeData(), DepartmentData(), ShopData())
        
        for store in Shop.select():
            if store.id == 1:
                assert (store.employee == 
                        Employee.get(self.fxt.nerd_employee.id))
                assert (store.department == 
                        Department.get(self.fxt.nerds_department.id))
            else:
                raise AssertionError('unexpected row: %s' % store)
    
    def test_register_on_load(self):
        
        ## data passed in via keyword, can't register that ...
        LoadMap.clear()
        self.fxt = SOFixture(   
                    data=(('foo',dict(name='foo',hair_color='red')),), 
                                so_class=Employee, clean=True)
        assert not LoadMap.is_registered(self.fxt)
        
        for emp in Employee.select():
            self.fxt.assert_has(emp, transkey={'hair_color':'hairColor'})
        
        self.fxt.clean()
        
        # this one uses meta data, and should be registered :
        self.fxt = EmployeeData()
        assert LoadMap.is_registered(self.fxt)